(************************************************)
(* Dynamic Link Libraries manager               *)
(* for TMT Pascal compiler 2.0+                 *)
(* Copyleft (C) 1998-2000 Virtual Research      *)
(* All lefts reserved                           *)
(*                                              *)
(* Unit version : 2.14                          *)
(* System       : DPMI32                        *)
(* Last updated : 11.Jun.2000                   *)
(************************************************)

unit DLL;
{.$DEFINE _DEBUG_}

interface

 const Version='2.14';

 type TExportTableHeader=record
  Signature:array[0..7] of char;
  Table:pointer;
  DummyString:pchar;
 end;
      TExportTableEntry=record
  Name:pchar;
  ParmCount:word;
  Address:pointer;
 end;
      PExportTableEntry=^TExportTableEntry;
      TDLLHandle=byte;

 var dllError:integer;
     dllExitCode:byte;
     dllSetES2PSP:boolean;

 function dllLoad(FileName:string):TDLLHandle;
 procedure dllFree(Handle:TDLLHandle);
 function dllGetDummyString(Handle:TDLLHandle):pchar;
 function dllCallByName(Handle:TDLLHandle;Name:pchar;Parms:pointer):dword;
 function dllCallByIndex(Handle:TDLLHandle;Index:dword;Parms:pointer):dword;
 function dllGetObjectAddressByName(Handle:TDLLHandle;Name:pchar):pointer;
 function dllGetObjectAddressByIndex(Handle:TDLLHandle;Index:dword):pointer;
 function dllGetVersion(Handle:TDLLHandle):string;

 const NO_ERROR=0;
       INVALID_FORMAT=1;
       NOT_ENOUGH_MEMORY=2;
       NO_MORE_HANDLES=3;
       FILE_NOT_FOUND=4;
       SYMBOL_NOT_FOUND=5;
       INDEX_OUT_OF_RANGE=6;
       INVALID_HANDLE=7;
       DPMI_ERROR=8;

       MAX_ERROR=8;

       dllErrorMessage:array[0..MAX_ERROR] of pchar=
       ('Operation successful',
        'Invalid DLL file format',
        'Not enough extended memory',
        'No more free DLL handles',
        'File not found',
        'Symbol not found in DLL',
        'Symbol index out of range',
        'Invalid DLL handle',
        'DPMI error');


implementation
{$I release.opt}

uses UnLZ;

{$I MZHEADER.INC}
{$I D32HEADR.INC}

const IMAGE_COMPRESSED=1;
      FIXUPS_COMPRESSED=4;

type  TDLLDescriptor=record
       BaseAddress:pointer;
       ExportTableAddress:pointer;
       BlockSize:dword;
       DummyString:pchar;
      end;


var LoadedDLLs:byte;
    DLLs:array[1..255] of TDLLDescriptor;

    OldInt21:farpointer;
    SaveSS:word;
    SaveESP:dword;

{$IFDEF _DEBUG_}
        (* We use two static 128 kb memory pools for debug purposes *)
    DLLPool:array[0..131071] of byte;
    FixupTablePool:array[0..131071] of byte;
{$ENDIF}

function ExecuteModule(EntryPoint:dword):byte;assembler;
(* Executes preloaded module *)
asm
 mov ax,204h            (* Get int21 vector *)
 mov bl,21h
 int 31h
 mov dword ptr OldInt21,edx
 mov word ptr OldInt21+4,cx
 mov ax,205h
 mov cx,cs
 mov edx,offset @@Hook21
 int 31h
 jc @@DPMIError
 push ebp               (* Save stack frame *)
 push ds
 push es                (* Save data selectors *)
 mov SaveSS,ss          (* Save stack selector *)
 mov SaveESP,esp        (* Save stack pointer *)

 cmp dllSetES2PSP,true
 jne @@DoNotSetES
 mov ah,62h
 int 21h
 mov es,bx
@@DoNotSetES:

 xor eax,eax
 xor ebx,ebx
 xor ecx,ecx
 xor edx,edx
 xor esi,esi
 xor edi,edi
 jmp EntryPoint

@@Hook21:
 cmp ah,4Ch
 je @@ModuleTerminated
 jmp cs:OldInt21

@@ModuleTerminated:
 mov ss,cs:SaveSS           (* Restore stack *)
 mov esp,cs:SaveESP
 pop es
 pop ds                     (* Restore data selectors *)
 pop ebp                    (* Restore stack frame *)
 and eax,0FFh
 push eax
 mov ax,205h                (* Restore int21 vector *)
 mov bl,21h
 mov edx,dword ptr OldInt21
 mov cx,word ptr OldInt21+4
 int 31h
 pop eax
 jnc @@Exit
@@DPMIError:
 mov dllError,DPMI_ERROR
@@Exit:
end;

function dllLoad(FileName:string):TDLLHandle;
(* Loads module into memory, applies fixups and 
   executes initialization section *)
var Module:file;
    StubHeader:TMZHeader;
    Header:TDOS32Header;
    Dummy,StubSize,FixupsSize:dword;
    i:byte;
    ProgramImage,FixupTable:pointer;

 function ExtractFileExt(Pathname:string):string;
 begin
  while Pos('\',Pathname)<>0 do
   Pathname:=Copy(Pathname,Pos('\',Pathname)+1,Length(Pathname)-Pos('\',Pathname));
  if Pos('.',Pathname)=0 then Result:='' else
   Result:=Copy(Pathname,Pos('.',Pathname)+1,Length(Pathname)-Pos('.',Pathname));
 end;

begin
 dllError:=NO_ERROR; dllExitCode:=0; Result:=0;
 if LoadedDLLs=255 then begin
  dllError:=NO_MORE_HANDLES;
  Exit;
 end;
 if ExtractFileExt(FileName)='' then FileName:=FileName+'.DLL';
 Assign(Module,FileName);
{$I-} 
 Reset(Module,1);
{$I+}
 if IOResult<>0 then begin
  dllError:=FILE_NOT_FOUND;
  Exit;
 end;
 BlockRead(Module,StubHeader,SizeOf(TMZHeader)); 
 if (StubHeader.Signature='MZ') or (StubHeader.Signature='ZM') then
  StubSize:=StubHeader.BytesOnLastPage+
            (dword(StubHeader.Pages-
             byte(StubHeader.BytesOnLastPage<>0))) shl 9
 else StubSize:=0;
 if StubSize=FileSize(Module) then begin
  Close(Module);
  dllError:=INVALID_FORMAT;
  Exit;
 end;
 Seek(Module,StubSize);
 BlockRead(Module,Header,SizeOf(TDOS32Header)); 
 if (Header.Signature<>'DLL ') or (Header.ImageOffset<44) then begin
  Close(Module);
  dllError:=INVALID_FORMAT;
  Exit;
 end;
{$IFNDEF _DEBUG_}
 if (MemAvail<(Header.MemoryRequired+32768)) or
    (MaxAvail<(Header.MemoryRequired)) then begin
  Close(Module);
  dllError:=NOT_ENOUGH_MEMORY;
  Exit;
 end;
{$ENDIF}
 for i:=1 to 255 do
  if DLLs[i].BaseAddress=nil then begin
   Result:=i; Break;
  end;
{$IFDEF _DEBUG_}
 DLLs[i].BaseAddress:=@DLLPool;
{$ELSE}
 GetMem(DLLs[i].BaseAddress,Header.MemoryRequired);
{$ENDIF}
 ProgramImage:=DLLs[i].BaseAddress;
 DLLs[i].BlockSize:=Header.MemoryRequired;
 Seek(Module,StubSize+Header.ImageOffset);
 if (Header.Flags and IMAGE_COMPRESSED=0) then
  BlockRead(Module,ProgramImage^,Header.ImageLength,Dummy)
 else 
  Header.ImageLength:=lzsDecode(Module,ProgramImage,Header.ImageLength);
 Seek(Module,StubSize+Header.StrippedLength);
 if (Header.Flags and FIXUPS_COMPRESSED=0) then
  FixupsSize:=Header.TMTRelocTableSize
 else
  BlockRead(Module,FixupsSize,4,Dummy);
{$IFNDEF _DEBUG_}
 if MaxAvail<FixupsSize then begin
  FreeMem(DLLs[i].BaseAddress,Header.MemoryRequired);
  DLLs[i].BaseAddress:=nil;
  Close(Module);
  dllError:=NOT_ENOUGH_MEMORY;
  Exit;
 end;
{$ENDIF}
{$IFDEF _DEBUG_}
 FixupTable:=@FixupTablePool;
{$ELSE}
 GetMem(FixupTable,FixupsSize);
{$ENDIF}
 if (Header.Flags and FIXUPS_COMPRESSED=0) then
  BlockRead(Module,FixupTable^,FixupsSize,Dummy)
 else
  Header.TMTRelocTableSize:=
   lzsDecode(Module,FixupTable,Header.TMTRelocTableSize);
 Close(Module);
 FillChar((ProgramImage+Header.ImageLength)^,    (* Initialize global data *)
          (Header.MemoryRequired-Header.ImageLength),0);    
 (* Apply fixups *)                                                
 asm
  mov edi,ProgramImage
  mov esi,FixupTable
  mov ecx,Header.TMTRelocTableSize
  shr ecx,2
@@1:
  lodsd
  add [edi+eax-4],edi
  loop @@1
 end;
{$IFNDEF _DEBUG_}
 FreeMem(FixupTable,FixupsSize);
{$ENDIF}
 DLLs[i].ExportTableAddress:=
  pointer(MemD[dword(Header.EIP+ProgramImage+8)]);
 Header.ESP+:=dword(ProgramImage);
 dllExitCode:=ExecuteModule(Header.ESP);
 if dllError=NO_ERROR then begin
  Inc(LoadedDLLs);
  DLLs[i].DummyString:=
   pointer(MemD[dword(Header.EIP+ProgramImage+12)]);
         (* Numeric constants suck.. these should be changed to
            SizeOf(pointer) etc.. maybe they will some day. *)
 end
 else begin
{$IFNDEF _DEBUG_}
  FreeMem(DLLs[i].BaseAddress,DLLs[i].BlockSize);
{$ENDIF}
  DLLs[i].BaseAddress:=nil;
  DLLs[i].ExportTableAddress:=nil;
  DLLs[i].BlockSize:=0;
 end;
end;

procedure dllFree(Handle:TDLLHandle);
(* Unloads DLL from memory and frees DLL handle *)
begin
 if DLLs[Handle].BaseAddress=nil then dllError:=INVALID_HANDLE
 else begin
  dllError:=NO_ERROR;
{$IFNDEF _DEBUG_}
  FreeMem(DLLs[Handle].BaseAddress,DLLs[Handle].BlockSize);
{$ENDIF}
  DLLs[Handle].BaseAddress:=nil;
  DLLs[Handle].ExportTableAddress:=nil;
  DLLs[Handle].BlockSize:=0;
  Dec(LoadedDLLs);
 end;
end;

function dllGetDummyString(Handle:TDLLHandle):pchar;
begin
 dllError:=NO_ERROR;
 if DLLs[Handle].BaseAddress=nil then dllError:=INVALID_HANDLE else
  Result:=DLLs[Handle].DummyString; (* Usually this string contains
                                       DLL description and/or copyright info *)
end;


function StrCmp(p1,p2:pchar):boolean;assembler;
(* Compares two null-terminated strings *)
asm
 mov edi,p1
 mov ebx,edi
 mov ecx,-1
 xor al,al
 repne scasb
 not ecx
 mov edi,ebx
 mov esi,p2
 repe cmpsb
 je @@Equal
 mov eax,0      (* not equal *)
 jmp @@Exit
@@Equal:
 mov eax,1
@@Exit:
end;

function dllCallByName(Handle:TDLLHandle;Name:pchar;Parms:pointer):dword;
(* Calls imported function by name *)
var Entry:PExportTableEntry;
    ParmCount:word;
    CallAddress:pointer;
begin
 dllError:=NO_ERROR;
 if DLLs[Handle].BaseAddress=nil then begin
  dllError:=INVALID_HANDLE;
  Exit;
 end;
 Entry:=DLLs[Handle].ExportTableAddress;
 while (not StrCmp(Entry^.Name,Name)) and (Entry^.Name<>nil) do 
  Inc(Entry,SizeOf(TExportTableEntry));
 if Entry^.Name=nil then begin
  dllError:=SYMBOL_NOT_FOUND;
  Exit;
 end;
 ParmCount:=Entry^.ParmCount;
 CallAddress:=Entry^.Address;
 asm
  mov esi,Parms
  movzx ecx,ParmCount
  jecxz @@Call
@@1:
  push dword ptr [esi]
  add esi,4
  loop @@1
@@Call:
  call [CallAddress]
  mov Result,eax
 end;
end;

function dllCallByIndex(Handle:TDLLHandle;Index:dword;Parms:pointer):dword;
(* Calls imported function by index in export table *)
var Entry:PExportTableEntry;
    ParmCount:word;
    CallAddress:pointer;
    I:dword;
begin
 dllError:=NO_ERROR;
 if DLLs[Handle].BaseAddress=nil then begin
  dllError:=INVALID_HANDLE;
  Exit;
 end;
 Entry:=DLLs[Handle].ExportTableAddress; I:=0;
 while (I<>Index) and (Entry^.Name<>nil) do begin
  Inc(Entry,SizeOf(TExportTableEntry)); Inc(I);
 end;
 if Entry^.Name=nil then begin
  dllError:=INDEX_OUT_OF_RANGE;
  Exit;
 end;
 ParmCount:=Entry^.ParmCount;
 CallAddress:=Entry^.Address;
 asm
  mov esi,Parms
  movzx ecx,ParmCount
  jecxz @@Call
@@1:
  push dword ptr [esi]
  add esi,4
  loop @@1
@@Call:
  call [CallAddress]
  mov Result,eax
 end;
end;

function dllGetObjectAddressByName(Handle:TDLLHandle;Name:pchar):pointer;
(* Gets object address from export table by its name *)
var Entry:PExportTableEntry;
begin
 dllError:=NO_ERROR;
 if DLLs[Handle].BaseAddress=nil then begin
  dllError:=INVALID_HANDLE;
  Exit;
 end;
 Entry:=DLLs[Handle].ExportTableAddress;
 while (not StrCmp(Entry^.Name,Name)) and (Entry^.Name<>nil) do
  Inc(Entry,SizeOf(TExportTableEntry));
 if Entry^.Name=nil then begin
  dllError:=SYMBOL_NOT_FOUND;
  Exit;
 end;
 Result:=Entry^.Address;
end;

function dllGetObjectAddressByIndex(Handle:TDLLHandle;Index:dword):pointer;
(* Gets object address by its index in export table *)
var Entry:PExportTableEntry;
    I:dword;
begin
 dllError:=NO_ERROR;
 if DLLs[Handle].BaseAddress=nil then begin
  dllError:=INVALID_HANDLE;
  Exit;
 end;
 Entry:=DLLs[Handle].ExportTableAddress; I:=0;
 while (I<>Index) and (Entry^.Name<>nil) do begin
  Inc(Entry,SizeOf(TExportTableEntry)); Inc(I);
 end;
 if Entry^.Name=nil then begin
  dllError:=INDEX_OUT_OF_RANGE;
  Exit;
 end;
 Result:=Entry^.Address;
end;

function dllGetVersion(Handle:TDLLHandle):string;
    (* This function translates VersionInfo constant into readable string *)
var VersionInfo:^longint;
    VersionByte:byte;
    VersionStr:string;
begin
 Result:='';
 VersionInfo:=dllGetObjectAddressByName(Handle,'VersionInfo');
 if (dllError=NO_ERROR) and (VersionInfo<>nil) then begin
  VersionByte:=(VersionInfo^ and $00FF0000) shr 16;    (* Major number *) 
  Str(VersionByte,Result);
  VersionByte:=(VersionInfo^ and $0000FF00) shr 8;     (* Minor number *)
  Str(VersionByte,VersionStr);
  Result:=Result+'.'+VersionStr;
  VersionByte:=(VersionInfo^ and $000000FF);           (* Revision *)
  if VersionByte in [97..122] then Result:=Result+Chr(VersionByte) else begin
   Str(VersionByte,VersionStr);
   Result:=Result+VersionStr;
  end;
  VersionByte:=(VersionInfo^ and $FF000000) shr 24;     (* Build *)
  Str(VersionByte,VersionStr);
  Result:=Result+'.'+VersionStr;
 end;
end;


begin
 for LoadedDLLs:=1 to 255 do begin
   DLLs[LoadedDLLs].BaseAddress:=nil;
   DLLs[LoadedDLLs].ExportTableAddress:=nil;
   DLLs[LoadedDLLs].BlockSize:=0;
 end;
 LoadedDLLs:=0;
 dllSetES2PSP:=false;
end.

